/*
 * barrier.cpp
 *
 *  Created on: Mar 16, 2024
 *      Author: Dario
 */

#include <sta/rtos/patterns/barrier.hpp>
#include <sta/rtos/thread.hpp>

#include <sta/time.hpp>
#include <sta/debug/assert.hpp>


namespace sta
{
	Barrier::Barrier(const char * name, uint32_t waitTime /* = 0 */)
		: mutex_{name},
		  flag_{},
		  waitTime_{waitTime},
		  lastTime_{now()},
		  threads_{},
		  nThreads_{0},
		  nEntered_{0},
		  nLeft_{0}
	{
		STA_ASSERT(name != nullptr);
	}

	void Barrier::subscribe()
	{
		mutex_.acquire();
		threads_.insert(RtosThread::getCurrentlyRunningThreadID());
		nThreads_ += 1;
		nLeft_ += 1;
		mutex_.release();
	}

	uint8_t Barrier::getNumSubscribedThreads()
	{
		mutex_.acquire();
		uint8_t rslt = nThreads_;
		mutex_.release();

		return rslt;
	}

	uint8_t Barrier::getNumEntered()
	{
		mutex_.acquire();
		uint8_t rslt = nEntered_;
		mutex_.release();

		return rslt;
	}

	/**
	 * 
	 * @return uint8_t Returns the number of threads that have left the barrier so far.
	 */
	uint8_t Barrier::getNumLeft()
	{
		mutex_.acquire();
		uint8_t rslt = nLeft_;
		mutex_.release();

		return rslt;
	}

	void Barrier::setMinimumWaitDuration(uint32_t duration)
	{
		mutex_.acquire();
		waitTime_ = duration;
		mutex_.release();
	}

	void Barrier::wait()
	{
		/**
		 * Taken from: https://en.wikipedia.org/wiki/Barrier_(computer_science)#Implementation
		 */
		mutex_.acquire();

		// Make sure the current thread is actually subscribed to this barrier.
		STA_ASSERT(threads_.find(RtosThread::getCurrentlyRunningThreadID()) != threads_.end());

		// Is this the first thread entering the barrier?
		if (nEntered_ == 0)
		{
			if (nLeft_ == nThreads_)
			{
				// First thread to arrive clears the flag.
				flag_.clear();
			}
			else
			{
				// This branch covers the case where a thread enters this barrier while others are still leaving.
				// This can happen when a thread is much faster than the others.
				mutex_.release();

				// Busy waiting until the last thread leaves.
				while (nLeft_ != nThreads_)
				{
					osThreadYield();
				}

				mutex_.acquire();

				// Clear the flag to make sure that entering threads cannot pass.
				flag_.clear();
			}
		}

		// Register this thread entering the barrier.
		++nEntered_;

		if (nEntered_ == nThreads_)
		{
			if (now() - lastTime_ < waitTime_)
			{
				// Let the thread sleep for the remaining time.
				osDelay(waitTime_ - (now() - lastTime_));
			}

			// If all threads are waiting in the barrier, set the flag allowing threads to leave.
			nEntered_ = 0;
			nLeft_ = 1;

			mutex_.release();
			flag_.set();
		}
		else
		{
			mutex_.release();

			// Wait in the barrier until the flag is set.
			flag_.wait();

			// Register that this thread left the barrier.
			mutex_.acquire();
			nLeft_++;
			mutex_.release();
		}
	}

	void Barrier::externalWait()
	{
		flag_.wait();
	}
}  // namespace sta


